from vaderSentiment.vaderSentiment import SentimentIntensityAnalyzer
import json
from .utils_funcs import text_iterate


SIA = SentimentIntensityAnalyzer()
r_arg = {'pct':True,'method':'dense'}

################################ CHECK THESE THRESHOLDS ####################################
def modify_vader(SS):
    try: 
        df = SS.polarities
    except:
        SS.create_df()
        df = SS.polarities
    
    ## Filter DF
    df_pos, df_neg, remove = filter_df(df, SS.NEU_THRESH, SS.VAR_UPPER, SS.VAR_LOWER)
    return make_VADER_custom(df_pos, df_neg, remove)

    
def make_VADER_custom(positive, negative, remove):
    SIA_Custom = SentimentIntensityAnalyzer()
    
    for i in remove: 
        try: del SIA_Custom.lexicon[i]
        except: pass
    
    for i in positive.itertuples():
        SIA_Custom.lexicon[i.Index] = i.polarity
        if i.Index in SIA_Custom.emojis:
            del SIA_Custom.emojis[i.Index]

    for i in negative.itertuples():
        SIA_Custom.lexicon[i.Index] = i.polarity
        if i.Index in SIA_Custom.emojis:
            del SIA_Custom.emojis[i.Index]
            
    return SIA_Custom

def filter_df(df,neu_thresh, var_upper, var_lower):

    df_remove = df[ (df.pos_prox.rank(**r_arg) < neu_thresh) 
                     & (df.neg_prox.rank(**r_arg) < neu_thresh)].index.tolist()
    
    df_k = df[~df.index.isin(df_remove)].copy()

    df_pos = df_k[(df_k.pos_prox.rank(**r_arg)>var_upper) & (df_k.neg_prox.rank(**r_arg)<var_lower)].copy()
    df_neg = df_k[(df_k.neg_prox.rank(**r_arg)>var_upper) & (df_k.pos_prox.rank(**r_arg)<var_lower)].copy()

    return df_pos, df_neg, df_remove


def apply_vader(self,save_outputs,return_outputs):
    if not self.classify:
        raise ValueError(f"""Apply model.fit() before model.transform()""")
    
    results = []
    
    if save_outputs:
        if self.VERBOSE: print(f"Saving Classified Text to: {self.paths['output_pols']}")
        with open(self.paths['output_pols'],'w') as newfile:
            for row in text_iterate(self, show=self.VERBOSE):
                result = [row, self.classify.polarity_scores(row)]
                output = json.dumps({'body':result[0],'polarity':result[1]})
                newfile.write(output+'\n')
                
                if return_outputs:
                    results.append(result)

    elif return_outputs:
        if self.VERBOSE: print('Returning Classified Text')
        for row in text_iterate(self, show=self.VERBOSE):
            result = [row, self.classify.polarity_scores(row)]
            results.append(result)

    if return_outputs:
        return results



import pandas as pd
from vaderSentiment.vaderSentiment import SentimentIntensityAnalyzer
r_arg = {'pct':True,'method':'dense'}

"""
Seed words for propagating polarity scores.
"""
POSITIVE_DERIVED_TWITTER = [
    'lovely', 'excellent', 'fortunate', 'pleasant', 'delightful',
    'perfect', 'loved', 'love', 'good', 'nice', 'beautiful', 'great',
    'enjoy', 'gorgeous', 'awesome', 'amazing', 'excited','loves','♥','🙂'
]

NEGATIVE_DERIVED_TWITTER = [
    'bad', 'horrible', 'hate', 'damn', '🙁', 'shit', 'shitty', 'fuck', 'hell',
    'wtf', 'hated','stupid', 'terrible', 'nasty', 'awful', 'worst', 'crap','crappy',
    'sad','bitch','hates'
]

POSITIVE_TURNEY = [ "good", "nice", "excellent", "positive", "fortunate", "correct", "superior"]
NEGATIVE_TURNEY = [ "bad","terrible","poor","negative","unfortunate","wrong","inferior"]

POSITIVE_FINANCE = ["successful","excellent","profit","beneficial","improving","improved","success","gains","positive"]
NEGATIVE_FINANCE = ["negligent","loss","volatile","wrong","losses","damages","bad","litigation","failure","down","negative"]

POSITIVE_TWEET = [ "love","loved","loves","awesome","nice","amazing","best","fantastic","correct","happy"]
NEGATIVE_TWEET = [ "hate","hated","hates","terrible","nasty","awful","worst","horrible","wrong","sad"]

MALE = ["man", "male", "boy", "gentleman", "mr", "masculine", "dad", "father", "brother", "son"]
FEMALE = ["female", "woman", "girl", "lady", "mom", "mum", "sister", "mother", "feminine", "daughter", "ms", "mrs", "miss"]

POSITIVE_HIST = [ "good","lovely","excellent","fortunate","pleasant","delightful","perfect","loved","love","happy"]
NEGATIVE_HIST = [ "bad","horrible","poor","unfortunate","unpleasant","disgusting","evil","hated","hate","unhappy"]

seed_dict = {'derived_twitter': [POSITIVE_DERIVED_TWITTER, NEGATIVE_DERIVED_TWITTER],
    'historical': [POSITIVE_HIST, NEGATIVE_HIST],
    'finance': [POSITIVE_FINANCE, NEGATIVE_FINANCE],
    'twitter': [POSITIVE_TWEET, NEGATIVE_TWEET],
    'gender': [MALE, FEMALE],
    'truney': [POSITIVE_TURNEY, NEGATIVE_TURNEY],
    
}


    
def set_seeds(SS):
    'returns seed options, or filters given seeds'
    
    try:
        if SS.POS_SEEDS:
            seeds = [SS.POS_SEEDS, SS.NEG_SEEDS]
        elif SS.PREDEFINED_SEEDS:
            seeds = seed_dict[SS.PREDEFINED_SEEDS.lower()]
            

        if type(seeds[0]) == list:
            pos_seeds = [SS.clean_text(i)[0] for i in sorted(set(seeds[0]))]
            neg_seeds = [SS.clean_text(i)[0] for i in sorted(set(seeds[1]))]
            return [pos_seeds,neg_seeds]

        if type(seeds[0]) == dict:
            pos_seeds = {SS.clean_text(i)[0]: seeds[0][i] for i in sorted(seeds[0])}
            neg_seeds = {SS.clean_text(i)[0]: seeds[1][i] for i in sorted(seeds[1])}
            return [pos_seeds,neg_seeds]
    except:
        valid_options = ', '.join(list(seed_dict))
        raise ValueError(f"""\nInvalid seed type. Please use ONE of the following options: {valid_options}.
                         Or supply your own in the following format: [upper_list_of_seeds, lower_list_of_seeds]""")



def custom_seeds(SS, pos_initial, neg_initial, n=10, return_all=False, sentiment = True):
    try:
        SS.load('PPMI')
    except:
        SS.gen_only()
        
    gdict = SS.load('dict')
    token2id = gdict.token2id    
    ppmi = SS.load('ppmi')
    
    pos_loc = [token2id[i] for i in pos_initial if i in token2id]
    neg_loc = [token2id[i] for i in neg_initial if i in token2id]

    pos_score = ppmi[pos_loc].toarray().mean(axis=0)
    neg_score  = ppmi[neg_loc].toarray().mean(axis=0)
    df_ppmi = pd.DataFrame({'upper_seeds':pos_score, 'lower_seeds':neg_score},index = token2id)
    df_ppmi['counts'] = [gdict.cfs[token2id[i]] for i in token2id]


    if sentiment:
        SIA = SentimentIntensityAnalyzer()
        df_ppmi['Vader'] = [SIA.lexicon[x] if x in SIA.lexicon else 0 for x in token2id]
        df_ppmi = df_ppmi[df_ppmi['Vader']!=0].copy()
        df_ppmi['r_diff'] = df_ppmi.rank(**r_arg).diff(axis=1)['lower_seeds']*-1
        df_ppmi['compound_seed_metric'] = (df_ppmi.counts*df_ppmi.Vader).rank(**r_arg)*0.5
        df_ppmi['compound_seed_metric'] += df_ppmi.r_diff.rank(**r_arg)
    else:
        df_ppmi['r_diff'] = df_ppmi.rank(**r_arg).diff(axis=1)['lower_seeds']*-1
        df_ppmi['compound_seed_metric'] = (df_ppmi.counts*df_ppmi.r_diff/df_ppmi.r_diff.abs()).rank(**r_arg)*0.5 
        df_ppmi['compound_seed_metric'] += df_ppmi.r_diff.rank(**r_arg)

    df_ppmi['compound_seed_metric'] = (df_ppmi['compound_seed_metric']*2/1.5) - 1
    df_ppmi = df_ppmi.sort_values('compound_seed_metric')
    
    if return_all:
        return df_ppmi
    
    df_ppmi = df_ppmi[~df_ppmi.index.isin(pos_initial+neg_initial)]
    pos_return = df_ppmi.tail(n).index.tolist()
    neg_return = df_ppmi.head(n).index.tolist()
    
    return [pos_return, neg_return]
    



import numpy as np
from collections import Counter
from sklearn.utils.extmath import randomized_svd
from scipy.sparse.linalg import svds
from scipy.sparse import diags, lil_matrix
from gensim.matutils import corpus2csc
from gensim.corpora import Dictionary
from tqdm.auto import tqdm
from .utils_funcs import text_iterate




def embed_text(SS):

    ### Load Tweets
    gdict = Dictionary()
    for row in text_iterate(SS, show=SS.VERBOSE):
        
        processed_comment = SS.clean_text(row)
        gdict.add_documents([processed_comment],prune_at = None)

    ### Filter Dictionary
    pos_seeds, neg_seeds = SS.SEEDS
    keep = list(SS.KEEP)

    keeptokens = list(pos_seeds)+ list(neg_seeds) + keep
    gdict.filter_extremes(no_above=SS.NO_ABOVE_1, no_below=SS.NO_BELOW, keep_tokens = keeptokens)
    gdict.compactify()

    ### Coocurrence
    if SS.VERBOSE: print('Running Cooc')
    cooc = gen_cooc(gdict, SS)
    
    SS.save( fname = SS.OUTPUT + 'cooc.pkl', var = cooc)

    ### PPMI   
    if SS.VERBOSE: print('Running PPMI')
    ppmi = gen_PPMI(cooc,SS)

    SS.save( fname = SS.OUTPUT + 'ppmi.pkl', var = ppmi)
    SS.save( fname = SS.OUTPUT + 'ppmi_index.pkl', var = gdict.token2id)

    ### SVD
    if SS.VERBOSE: print('Making Low Dim')
    u,s, v = run_lowdim(ppmi,SS)


    np.save(SS.OUTPUT + "vec-u.npy", u)
    np.save(SS.OUTPUT + "vec-v.npy", v)
    np.save(SS.OUTPUT + "vec-s.npy", s)
    
    SS.save( fname = SS.OUTPUT + 'dict.pkl', var = gdict)

def gen_cooc(gdict,SS, tfidf = False):
    bow_corpus = []
    dup_corpus = Counter()


    for row in text_iterate(SS, show=SS.VERBOSE):
        processed_comment = SS.clean_text(row)
        mat = gdict.doc2bow(processed_comment)
        bow_corpus += [mat]
        dup_corpus.update(dict(mat)) # To subtract diagonal 

    if tfidf:
        return bow_corpus
        
    term_doc_mat = corpus2csc(bow_corpus)
    cooc = np.dot(term_doc_mat, term_doc_mat.T)
    diag = diags([dup_corpus[i] for i in sorted(dup_corpus)])

    cooc -= diag
    return cooc

def gen_PPMI(sparse_matrix, SS, cds = True):
    total_sum = sparse_matrix.sum()
    row_sum = np.array(sparse_matrix.sum(axis=1))
    col_sum = np.array(sparse_matrix.sum(axis=0))
    if cds:
        col_sum = col_sum**(0.75)

    row_sum = row_sum / row_sum.sum()
    col_sum = col_sum / col_sum.sum()
    numer_mat = (sparse_matrix / total_sum)
    pmi = lil_matrix(numer_mat.shape)


    for ind, numer in tqdm(enumerate(numer_mat.T), total=len(row_sum), disable=(not SS.VERBOSE)):
        numer = numer.toarray()
        denom = np.multiply(row_sum[ind], col_sum.copy())
        denom[numer == 0] = 1
        ppmi_row = numer / denom
        ppmi_row[ppmi_row == 0] = 1

        pmi[ind, :] = np.maximum(np.log(ppmi_row), np.zeros(len(numer)))

    return pmi.tocsr()



def run_lowdim(ppmi,SS, dim=300):
    try:
        u, s, v = svds(ppmi,k=dim, maxiter = 5, solver ='arpack', random_state=0)
    except:
        u, s, v = randomized_svd(ppmi, n_components=dim, n_iter=5,random_state=None)
        if SS.VERBOSE: print('RandomSVD')
    return u,s, v


    




stoplist = {
                'a', 'about', 'above', 'after', 'again', 'against', 'ain', 'all', 'am',
                'an', 'and', 'any', 'are', 'aren', "aren't", 'as', 'at', 'be', 'because',
                'been', 'before', 'being', 'below', 'between', 'both', 'but', 'by', 'can',
                'couldn', "couldn't", 'd', 'did', 'didn', "didn't", 'do', 'does', 'doesn',
                "doesn't", 'doing', 'don', "don't", 'down', 'during', 'each', 'few', 'for',
                'from', 'further', 'had', 'hadn', "hadn't", 'has', 'hasn', "hasn't",
                'have', 'haven', "haven't", 'having', 'he', 'her', 'here', 'hers',
                'herself', 'him', 'himself', 'his', 'how', 'i', 'if', 'im', 'in', 'into',
                'is', 'isn', "isn't", 'it', "it's", 'its', 'itself', 'just', 'll', 'm',
                'ma', 'me', 'mightn', "mightn't", 'more', 'most', 'mustn', "mustn't", 'my',
                'myself', 'needn', "needn't", 'no', 'nor', 'not', 'now', 'o', 'of', 'off',
                'on', 'once', 'only', 'or', 'other', 'our', 'ours', 'ourselves', 'out',
                'over', 'own', 're', 's', 'same', 'shan', "shan't", 'she', "she's",
                'should', "should've", 'shouldn', "shouldn't", 'so', 'some', 'such', 't',
                'than', 'that', "that'll", 'the', 'their', 'theirs', 'them', 'themselves',
                'then', 'there', 'these', 'they', 'this', 'those', 'through', 'to', 'too',
                'under', 'until', 'up', 've', 'very', 'was', 'wasn', "wasn't", 'we',
                'were', 'weren', "weren't", 'what', 'when', 'where', 'which', 'while',
                'who', 'whom', 'why', 'will', 'with', 'won', "won't", 'wouldn', "wouldn't",
                'y', 'you', "you'd", "you'll", "you're", "you've", 'your', 'yours',
                'yourself', 'yourselves'
            }


from tqdm.auto import tqdm
import csv

from string import punctuation
translator = str.maketrans('', '', punctuation)


def default_clean_text(SS, doc):
    "Input doc and return clean list of tokens"
    doc = doc.replace('\r', ' ').replace('\n', ' ')
    lower = doc.lower()  # all lower case
    nopunc = lower.translate(translator)  # remove punctuation
    words = nopunc.split()  # split into tokens
    if SS.STOPWORDS:
        words = [
            w for w in words
            if w not in SS.STOPWORDS and not w.isdigit()
        ]  # remove stopwords
    return words


def text_iterate(SS, show=False, full=False):
    'Allows iteration over a list of text or a jsonl/json file'
    if isinstance(SS.FILEINPUT, list):
        for row in tqdm(SS.FILEINPUT, total=SS.LINES,
                        disable=not show):
            yield row
    else:
        with open(SS.FILEINPUT, encoding='utf-8') as f:
            reader = csv.reader(f)
            for row in tqdm(reader, total=SS.LINES, disable=not show):
                if full:
                    yield row
                else:
                    yield ' ,'.join(row)



from .suggest_seeds import set_seeds, custom_seeds, seed_dict
from .stopwords import stoplist
from . import load_polarities
from . import create_embeddings
from . import run_bootstrapping
from . import create_vader
from . import utils_funcs
import warnings
import pickle
import os









class CIDER:

    def __init__(self,
                 fileinput,
                 output,
                 pos_seeds=None,
                 neg_seeds=None,
                 predefined_seeds='derived_twitter',
                 max_polarities_returned=5000,
                 preprocessing = 'default',
                 stopwords='nltk',
                 iterations=100,
                 verbose=True,
                 no_below=100,
                 keep=[]):
        '''
        Create domain specific lexicons.

        The code can be run as follows:
            cdr = CIDER(fileinput,output)
            cdr.fit_transform() 

        Parameters
        ----------
        fileinput - STRING/LIST - either list of STR or path of .json/.jsonl file using the format:
                                  {'text': example_text1}
                                  {'text': example_text2}
                                  {'text': example_text3}
                                    ... 
        output - STRING - path of folder for saving files
        pos_seeds - LIST/DICT - upper pole seed set. Either list, e.g. ['good','great'] 
                                                         or dict, e.g. {'good':1,'great':2}

        neg_seeds - LIST/DICT - lower pole seed set. Either list, e.g. ['bad','terrible'] 
                                                         or dict, e.g. {'bad':1,'terrible':2} (keep values +ve)

        predefined_seeds -STR - choose from 'derived_twitter', 'truney', 'finance', 'twitter', 'gender', 'historical'
        max_polarities_returned - INT - maximum number of words returned
        preprocessing - STR/FUNC/NONETYPE - set as 'default' for default preprocessing, None for text.split() or a 
                                            custom function taking one input (text) and returning a preprocessed and
                                            tokenised list of words/tokens.
        stopwords - STR/list/None - set as 'default' to remove nltk stopwords. Or add custom list/None. 
                                    ONLY WORKS WITH "preprocessing = 'default'"
        iterations - INT - the number of iterations for label propagation 

        verbose - BOOL - display progress
        no_below - INT - exclude words that occur less frequently than this
        keep - LIST/None - List of words to prevent being excluded
        '''
        if type(fileinput) == str:
            if not os.path.isfile(fileinput):
                raise FileNotFoundError('No file exists at the location specified')
        self.FILEINPUT = fileinput
        self.ITERATIONS = iterations
        if stopwords == 'default': stopwords = stoplist
        if stopwords == None: stopwords = []
        self.STOPWORDS = stopwords
        self.VERBOSE = verbose
        self.PREPROCESSING = preprocessing


        ## Seed Words
        self.POS_SEEDS = pos_seeds
        self.NEG_SEEDS = neg_seeds
        self.PREDEFINED_SEEDS = predefined_seeds
        self.SEEDS = set_seeds(self)

        ## Filtering Thresholds
        if keep == None: keep = []
        self.KEEP = sorted(set(self.clean_text(' '.join(keep))))
        self.NO_ABOVE_1 = 0.5
        self.NO_ABOVE_2 = 0.1
        self.NO_BELOW = no_below
        self.STATE = 0
        self.NN = 25
        self.LINES = 0
        self.MAX = max_polarities_returned

        ## Save Locations
        if not os.path.isdir(output):
            raise FileNotFoundError('No folder exists at the location specified')

        self.OUTPUT = output
        self.POLARITY_OUTPUT = f"{self.OUTPUT}polarities.json"

        self.paths = {
            'dict': f'{self.OUTPUT}dict.pkl',
            'ppmi': f'{self.OUTPUT}ppmi.pkl',
            'cooc': f'{self.OUTPUT}cooc.pkl',
            'ppmi_index': f'{self.OUTPUT}ppmi_index.pkl',
            'output_pols': f'{self.OUTPUT}output.json'
        }

        ## returned dataframe parameters
        self.CI = False
        self.STD = True
        self.return_all = False
        self.scale = True

        ## filtering polarities parameters. These can be adjusted.
        self.NEU_THRESH=0.45,
        self.VAR_UPPER=0.9,
        self.VAR_LOWER=0.5
        


    def generate_seeds(self, pos_initial, neg_initial, n=10, return_all=False, sentiment=True):
        '''
        Suggests Custom Seed words based off of PPMI relationships and frequency. These words are just
        suggestions and are not neccessarily the best for your data.
        
        Parameters
        ----------
        pos_initial - list - Small input list to base seeds off, e.g. ['amazing', 'brilliant', 'great']
        neg_initial - list - Small input list to base seeds off, e.g. ['terrible', 'rubbish', 'bad']
        n - INT - number of words returned in the produced pos & neg seed sets
        return_all - BOOL - set as True to return full filtered dataframe
        sentiment - BOOL - set as False if the generating non-sentiment seeds 
        '''
        return custom_seeds(self, pos_initial, neg_initial, n, return_all,sentiment)
    


    def fit(self, full_run=True):
        '''
        Train CIDER and fit a VADER model on the outputs

        Parameters
        ----------
        full_run - BOOL - set as False to just return previously executed results
        '''
        if full_run:
            ## Generating embeddings
            for ind, _ in enumerate(utils_funcs.text_iterate(self)):
                continue
            self.LINES = ind + 1
            if (self.LINES < 30000) & (self.NO_BELOW > 30):
                warnings.warn(f"\nThe size of the dataset ({self.LINES}) is relatively small. " 
                              "Consider reducing the 'no_below' parameter to improve returned " 
                              "polarities and to avoid potential errors.", stacklevel=2)

            create_embeddings.embed_text(self)

            ## Running Bootstrapping
            run_bootstrapping.propogate_labels(self)
        
        ## Loading Polarities
        self.create_df()
        self.train_VADER()



    def transform(self, save_outputs=False, return_outputs = True):
        '''
        Apply fitted VADER model on the inputs.

        Parameters
        ----------
        save_outputs - BOOL - save the outputs into .json file
        return_outputs - BOOL - return the outputs as a list
        '''
        return create_vader.apply_vader(self, save_outputs, return_outputs)
    


    def fit_transform(self, save_outputs=False, return_outputs = True, full_run = True):
        '''
        Train CIDER and fit a VADER model on the outputs. Apply fitted VADER model on the inputs
        
        Parameters
        ----------
        save_outputs - BOOL - save the outputs into .json file
        return_outputs - BOOL - return the outputs as a list
        full_run - BOOL - set as False to just return previously executed results

        '''
        self.fit(full_run)
        return self.transform(save_outputs, return_outputs)
    


    def create_df(self):
        'returns polarities only'
        self.polarities = load_polarities.make_df(self)



    def train_VADER(self):
        'trains VADER only'
        self.classify = create_vader.modify_vader(self)
        

    def clean_text(self,text):
        '''
        cleans and tokenises the text
        '''
        if self.PREPROCESSING == 'default':
            return utils_funcs.default_clean_text(self, text)
        if self.PREPROCESSING == None:
            return text.split()
        if callable(self.PREPROCESSING):
            return self.PREPROCESSING(text)

    def save(self, fname, var):
        'Function used to pickle (save) variables'
        with open(fname, "wb") as f:
            pickle.dump(var, f)



    def load(self, fname):
        'Loads pickled (saved) variables'
        if fname.lower() in self.paths:
            fname = self.paths[fname.lower()]

        with open(fname, "rb") as f:
            return pickle.load(f)


    
    def get_seeds(self,input = None):
        '''
        Either returns set seeds, or if input != None, the corresponding seed set for that input is returned 
        '''
        if input:
            return seed_dict[input.lower()]
        else:
            return self.SEEDS



import pandas as pd
from sklearn.preprocessing import StandardScaler
from vaderSentiment.vaderSentiment import SentimentIntensityAnalyzer
import scipy.stats as st
import numpy as np
import json
from collections import Counter
from collections import defaultdict


def make_df(SS):
    '''
    Makes df using data
    '''
    
    ### Collecting the Data
    pols = {'pos':defaultdict(list),'neg':defaultdict(list)}
    
    with open(SS.POLARITY_OUTPUT) as file:
        for row in file:
            boot_pols = json.loads(row)
            for w in boot_pols:
                if np.isnan(boot_pols[w][0]):
                    continue
                if np.isnan(boot_pols[w][1]):
                    continue
                pols['pos'][w].append(boot_pols[w][0])
                pols['neg'][w].append(boot_pols[w][1])                
                
    df = pd.DataFrame(pols)
    df['polarity'] = df.pos.apply(np.array)/(df.pos.apply(np.array)+df.neg.apply(np.array))
        
    ## Word Counts
    index = SS.load( fname = SS.OUTPUT + 'dict.pkl')
    word_freq = Counter({i[0]:i[1] for i in index.most_common()})
    df['freq'] = df.index.map(word_freq)     
        
    if SS.CI:
        df[f'CI_{SS.CI}'] = df['polarity'].apply(lambda x: st.norm.interval(confidence=0.95,scale=st.sem(x))[1])
        
    if SS.STD:
        df['polarity_std'] = df['polarity'].apply(np.std)
        
    
    ### Vader
    SIA = SentimentIntensityAnalyzer()
    df['VADER'] = df.index.to_series().apply(lambda x: SIA.lexicon[x.lower()] if x in SIA.lexicon else np.nan)


    if SS.return_all:
        return df
    
    df['pos_prox'] = df.pos.apply(np.mean)
    df['neg_prox'] = df.neg.apply(np.mean)
    df['polarity'] = df.polarity.apply(np.mean)
    df = df.drop(['pos','neg'],axis=1)
    
    ### Scaling the Data
    scaler = StandardScaler(with_std=False)
    df["polarity"] = scaler.fit_transform(df["polarity"].values.reshape(-1, 1))
        
    scale = df.polarity.abs().max()
    df['polarity'] = 4*df.polarity/scale
    df = df[['polarity','VADER','polarity_std','pos_prox','neg_prox','freq']]

    return df.sort_values('polarity')



from sklearn import preprocessing
from collections import Counter
from scipy import sparse
import numpy as np
from tqdm.auto import tqdm
import json


def propogate_labels(SS):

    ### Loading words to keep
    keep = list(SS.KEEP)
    pos_seeds, neg_seeds = SS.SEEDS
    keeptokens = list(pos_seeds)+ list(neg_seeds) + keep

    ### Filtering gdict
    gdict = SS.load(fname = SS.OUTPUT + 'dict.pkl')
    vocab = gdict.token2id
    gdict.filter_extremes(no_above=SS.NO_ABOVE_2, no_below=SS.NO_BELOW, keep_tokens = keeptokens)

    ## Filtering out SS.MAX tokens
    keeptoken_ids = [gdict.token2id[i] for i in keeptokens if i in gdict.token2id] 
    to_keep_init = [i[0] for i in Counter(gdict.dfs).most_common(SS.MAX)] # At most SS.MAX words
    to_keep = list(dict.fromkeys(keeptoken_ids+ to_keep_init))[:SS.MAX]
    to_keep = sorted(to_keep, key=lambda w: gdict.dfs[w], reverse=True)

    gdict.filter_tokens(good_ids=to_keep) # Remove Bad Tokens
    u = np.load(SS.OUTPUT + "vec-u.npy")

    sub_vec_m = preprocessing.normalize(u, copy=False)

    word_list = set(gdict.token2id.keys()).union(keeptokens)
    word_list =  set([w for w in word_list if w in vocab])

    keep_indices = [vocab[word] for word in word_list]
    sub_vec_m = sub_vec_m[keep_indices, :]

    params = {'POLARITY_OUTPUT': SS.POLARITY_OUTPUT,
              'ITERATIONS': SS.ITERATIONS,
              'VERBOSE': SS.VERBOSE,
              'STATE': SS.STATE,
              'SEEDS': SS.SEEDS}


    ### Transition Matrix
    trans_M = transition_matrix(sub_vec_m, SS.NN)

    polarity_bootstrap(trans_M, word_list, params)


def transition_matrix(embeddings,nn):
    """
    Build a probabilistic transition matrix from word embeddings.
    """
    L = similarity_matrix(embeddings,nn)

    Dinv = np.diag([1.0 / L[i].sum() for i in range(L.shape[0])])
    L = L.dot(Dinv)

    return L


def similarity_matrix(embeddings, nn, arccos=False, similarity_power=1):
    """
    Constructs a similarity matrix from embeddings.
    nn argument controls the degree.
    """
    
    def make_knn(vec, nn=nn):
        vec[vec < vec[np.argsort(vec)[-nn]]] = 0
        return vec

    L = embeddings.dot(embeddings.T)
    if sparse.issparse(L):
        L = L.todense()
    if arccos:
        L = np.arccos(np.clip(-L, -1, 1)) / np.pi
    else:
        L += 1
        
    np.fill_diagonal(L, 0)
    L = np.apply_along_axis(make_knn, 1, L, nn = nn)
    
    return L ** similarity_power

def polarity_bootstrap(M, wordlist,params):
    iterations = params['ITERATIONS']
    verbose = params['VERBOSE']
    state = params['STATE']
    saveloc = params['POLARITY_OUTPUT']
    
    np.random.seed(state)
    
    with open(saveloc,'w') as newfile:
        for _ in tqdm(range(iterations), disable=(not verbose)):
            [rpos, rneg] = run_random_walks(M, wordlist, params)

            pols = {w: [rpos[i], rneg[i]] for i, w in enumerate(wordlist)}
            json.dump(pols, newfile)
            newfile.write('\n')
            


def run_random_walks(M, wordlist, params):

    positive_seeds, negative_seeds = params['SEEDS']

    
    pos_len = len(positive_seeds) - 2
    neg_len = len(negative_seeds) - 2
    
    if pos_len > 10:  pos_len = int(len(positive_seeds)*0.8)
    if neg_len > 10:  neg_len = int(len(negative_seeds)*0.8)
    
    if pos_len <1: pos_len=1
    if neg_len <1: neg_len=1

    pos_seeds = np.random.choice(sorted(positive_seeds), pos_len, replace=False)
    neg_seeds = np.random.choice(sorted(negative_seeds), neg_len, replace=False)        
    
    if type(positive_seeds) == dict:
        pos_seeds = {i:positive_seeds[i] for i in pos_seeds}
        neg_seeds = {i:negative_seeds[i] for i in neg_seeds}
    
    
    polarities = random_walk(M, wordlist, pos_seeds, neg_seeds, beta=0.9)
    
    return polarities




def random_walk(M, words, positive_seeds, negative_seeds, beta=0.9):
    """
    Learns polarity scores via random walks with teleporation to seed sets.
    Main method used in paper. 
    """
        
    def run_random_walk(M, teleport, beta):
        def update_seeds(r):
            r += (1 - beta) * teleport / (np.sum(teleport)+1)

        return run_iterative(
            M * beta, np.ones(M.shape[1]) / M.shape[1], update_seeds)

    if not type(positive_seeds) is dict:
        positive_seeds = {word: 1.0 for word in positive_seeds}
        negative_seeds = {word: 1.0 for word in negative_seeds}
    
    

    rpos = run_random_walk(
        M, initialize_weighted_set(words, positive_seeds), beta)
    rneg = run_random_walk(
        M, initialize_weighted_set(words, negative_seeds), beta)
    

        
    return [rpos, rneg]

### HELPER METHODS #####
def initialize_weighted_set(words, seed_weights):
    return np.array(
        [seed_weights[word] if word in seed_weights else 0.0 for word in words]
    )


def run_iterative(M, r, update_seeds, max_iter=50, epsilon=1e-6):
    for _ in range(max_iter):
        last_r = np.array(r)
        r = np.dot(M, r)
        update_seeds(r)

        if np.abs(r - last_r).sum() < epsilon:
            break
    return r




